from abc import abstractmethod
from typing import Callable, Coroutine, Type
from arkitekt.messages.base import MessageModel
from enum import Enum
import asyncio
from herre import Herre
from herre.herre import get_current_herre
from koil import Koil
from koil.koil import get_current_koil
from fakts import Fakts, get_current_fakts, Config
import pydantic


class TransportConfig(Config):
    pass

class TansportConfigError(Exception):
    pass

class Transport(object):
    configClass = TransportConfig

    def __init__(self, config_dict, broadcast = None, herre: Herre = None, koil: Koil = None, fakts: Fakts = None, **kwargs) -> None:
        self._broadcast = broadcast
        assert self._broadcast is not None, "Please provide a broadcaster that receives messages"
        self.herre = herre or get_current_herre()
        self.koil = koil or get_current_koil()
        self.fakts = fakts or get_current_fakts()
        self.loop = self.koil.loop

        try:
            self.config = self.configClass(**config_dict)
        except pydantic.error_wrappers.ValidationError as e:
            raise TansportConfigError(f"Non valid Configuration {config_dict}") from e


    @abstractmethod
    def aconnect(self):
        raise NotImplementedError()


    @abstractmethod
    def adisconnect(self):
        raise NotImplementedError()

    async def broadcast(self, message: MessageModel):
        return await self._broadcast(message)


    @abstractmethod
    async def forward(self, message: MessageModel):
        raise NotImplementedError()






import asyncio
from arkitekt.config import TransportProtocol
from arkitekt.messages.base import MessageModel
from arkitekt.transport.registry import TransportRegistry, get_current_transport_registry
from arkitekt.transport.websocket import WebsocketTransport
from arkitekt.transport.base import Transport
from arkitekt.ward import ArkitektWard
from herre.herre import Herre, get_current_herre
from herre.wards.registry import get_ward_registry
from arkitekt.legacy.utils import *
from arkitekt.registry import set_current_rpc
from koil.koil import Koil, get_current_koil
from fakts import Fakts, get_current_fakts, Config

try:
    from rich.traceback import install
    install()
except:
    pass

import logging

logger = logging.getLogger(__name__)




class PostmanConfig(Config):
    type: TransportProtocol
    kwargs: dict

    class Config:
        group = "arkitekt.postman"


class Postman:

    def __init__(self, *args, auto_login=True, auto_connect=True, loop=None, register=True, herre: Herre = None, koil: Koil = None, fakts: Fakts = None, transport_registry: TransportRegistry = None, **kwargs) -> None:

        self.auto_login = auto_login
        self.auto_connect = auto_connect
        self.transport: Transport = None
        self.herre = herre or get_current_herre() 
        self.koil = koil or get_current_koil()
        self.fakts = fakts or get_current_fakts()
        self.loop = self.koil.loop
        self.ward: ArkitektWard = get_ward_registry().get_ward_instance("arkitekt")
        self.transport_registry = transport_registry or get_current_transport_registry()

        self.connected = False
        self.queues = {}

        if register:
            set_current_rpc(self)


        super().__init__(*args, **kwargs)



    async def adisconnect(self):
        await self.ward.adisconnect()
        await self.transport.adisconnect()
        logger.info(f"Successfully disconnected Postman {self}")


    async def broadcast(self, message: MessageModel):
        if message.meta.reference in self.queues:
            logger.info(f"RPC Message Received")
            await self.queues[message.meta.reference].put(message)
        else:
            logger.error(f"Message without a reply channel {message}")


    async def stream_replies_to_queue(self, message: MessageModel):
        """ Creates a queue for this referenced message
        and forwards every message on a reply channel to the return queue

        Args:
            message (MessageModel): [description]

        Returns:
            asyncio.Queue: The Reply Queue
        """
        reply_channel = message.meta.reference
        reply_queue = asyncio.Queue()
        self.queues[reply_channel] = reply_queue
        await self.transport.forward(message)
        return reply_queue


    async def forward(self, message: MessageModel):
        await self.transport.forward(message)
        
        
    async def aconnect(self):
        if not self.fakts.loaded:
            await self.fakts.aload()

        if not self.herre.logged_in:
            await self.herre.alogin()

        if not self.ward.connected:
            assert self.auto_connect, "We have not connected to Arkitekt before and autoconnect was set to false. Please connect a ward before or set auto_connect to True"
            await self.ward.aconnect()


        self.config = PostmanConfig.from_fakts(fakts=self.fakts)
        self.transcript = self.ward.transcript
        self.transport = self.transport_registry.get_transport_for_protocol(self.config.type)(self.config.kwargs, broadcast=self.broadcast)

        await self.transport.aconnect()
        self.connected = True


    async def __aenter__(self):
        await self.aconnect()
        return self


            
    async def __aexit__(self, *args, **kwargs):
        await self.adisconnect()









from herre.wards.graphql import ParsedQuery
from arkitekt.graphql.node import DETAIL_NODE_FR


TEMPLATE_GET_QUERY = ParsedQuery("""
query Template($id: ID,){
  template(id: $id){
    id
    node {
        """
        + DETAIL_NODE_FR+
        """
    }
  }
}
""")


UPDATE_OR_CREATE_TEMPLATE = ParsedQuery("""
mutation CreateTemplate($node: ID!, $params: GenericScalar, $extensions: [String], $version: String){
  createTemplate(node: $node, params: $params, extensions: $extensions, version: $version){
    id
    node {
        """
        + DETAIL_NODE_FR+
        """
    }
  }
}
""")




from arkitekt.agents.base import Agent
from arkitekt.messages.postman.provide.provide_log import ProvideLogMessage
from arkitekt.messages.postman.log import LogLevel
from arkitekt.monitor.monitor import AgentPanel, get_current_monitor
from herre.console.context import get_current_console
from arkitekt.packers.utils import shrink_outputs
from re import template
from arkitekt.messages.postman.assign.assign_cancelled import AssignCancelledMessage
from arkitekt.messages.postman.unassign.bounced_forwarded_unassign import BouncedForwardedUnassignMessage
from arkitekt.messages.postman.assign.assign_critical import AssignCriticalMessage
from typing import Dict
from arkitekt.messages.postman.assign.assign_return import AssignReturnMessage
from arkitekt.messages.postman.assign.bounced_forwarded_assign import BouncedForwardedAssignMessage
from arkitekt.messages.postman.assign.bounced_assign import BouncedAssignMessage
from arkitekt.messages.base import MessageModel
from arkitekt.messages.postman.provide.provide_transition import ProvideState
from arkitekt.transport.base import Transport
from arkitekt.packers.transpilers.base import Transpiler
from arkitekt.messages.postman.provide.bounced_provide import BouncedProvideMessage
import asyncio
from asyncio.tasks import Task
from arkitekt.messages.postman.provide import ProvideTransitionMessage
import logging
from arkitekt.schema.template import Template
from koil.koil import Koil, get_current_koil
from koil.loop import koil



logger = logging.getLogger(__name__)



class Actor:
    template: Template

    def __init__(self, *args, koil: Koil = None, strict=False, expand_inputs=True, shrink_outputs=True, transpilers= {}, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.koil = koil or get_current_koil()
        self.loop = self.koil.loop
        self.strict = strict
        self.expand_inputs = expand_inputs
        self.shrink_outputs = shrink_outputs
        self.runningAssignments: Dict[str, asyncio.Task] = {} # Running assignments indexed by assignment reference


    async def log(self, message, level = LogLevel.INFO):
        logger.info("{level}: {message}")


    async def provide_log(self, message: str, level = LogLevel.INFO):
        await self.agent.transport.forward(ProvideLogMessage(
                data= {
                    "level": level,
                    "message": message
                },
                meta=self.provision.meta.dict(exclude={"type"})
            ))
        logger.info(f"Provide Log: {level}: {message}")


    def run(self, *args, **kwargs):
        return koil(self.arun(*args, **kwargs))

    async def acall(self, message: MessageModel):
        await self.in_queue.put(message)


    async def on_provide(self, message: BouncedProvideMessage):
        print("Getting Called Here")
        return None


    async def on_unprovide(self, message: BouncedProvideMessage):
        return None


    async def on_assign(self, message: BouncedForwardedAssignMessage):
        raise NotImplementedError("Needs to be owerwritten in Actor Subclass")


    async def arun(self, provision: BouncedProvideMessage, agent: Agent):
        self.provision = provision
        self.agent = agent
        self.transport = agent.transport
        self.in_queue = asyncio.Queue()

        try:
            self.template = await Template.asyncs.get(id=self.provision.data.template)
            

            await self.transport.forward(ProvideTransitionMessage(
                data= {
                    "state": ProvideState.PROVIDING,
                    "message": "We just got started Bay"
                },
                meta={
                    "reference": self.provision.meta.reference,
                    "extensions": self.provision.meta.extensions
                }
            ))



            await self.on_provide(self.provision)

            await self.transport.forward(ProvideTransitionMessage(
                data= {
                    "state": ProvideState.ACTIVE,
                    "message": "We just got started Bay"
                },
                meta={
                    "reference": self.provision.meta.reference,
                    "extensions": self.provision.meta.extensions
                }
            ))

            while True:
                await self.log("Waiting for assignmements")
                message = await self.in_queue.get()
                logger.info(f"Received Message {message}")
                
                if isinstance(message, BouncedForwardedAssignMessage):
                    await self.log("Assigningment received")
                    task = self.loop.create_task(self.on_assign(message))
                    self.runningAssignments[message.meta.reference] = task

                if isinstance(message, BouncedForwardedUnassignMessage):
                    if message.data.assignation in self.runningAssignments:
                        task = self.runningAssignments[message.data.assignation]
                        if not task.done():
                            logger.info("Task is being cancelled")
                            task.cancel()
                        else:
                            logger.error("Task was already done")
                    else:
                        logger.error("Task was never assigned to this actor")
                        if self.strict: raise Exception("Received cancellation for Task that was never assinged to this actor!")
                        await self.transport.forward(AssignCancelledMessage(data={
                            "canceller": str("Cancelled because actor receiving this cancellation never had this task but was also not strict")
                        }, meta = {
                            "reference": message.data.assignation,
                            "extensions": message.meta.extensions
                        }))



        except Exception as e:
            logger.exception(e)
            await self.log(f"Provision Exception {str(e)}")
            await self.transport.forward(ProvideTransitionMessage(
                data= {
                    "state": ProvideState.CRITICAL,
                    "message": f"{e}"
                },
                meta={
                    "reference": self.provision.meta.reference,
                    "extensions": self.provision.meta.extensions
                }
            ))


        except asyncio.CancelledError as e:
            await self.transport.forward(ProvideTransitionMessage(
                data= {
                    "state": ProvideState.CANCELING,
                    "message": f"{e}"
                },
                meta={
                    "reference": self.provision.meta.reference,
                    "extensions": self.provision.meta.extensions
                }
            ))

            await self.on_unprovide(self.provision)


            logger.info("Doing Whatever needs to be done to cancel!")
            await self.transport.forward(ProvideTransitionMessage(
                data= {
                    "state": ProvideState.CANCELLED,
                    "message": f"{e}"
                },
                meta={
                    "reference": self.provision.meta.reference,
                    "extensions": self.provision.meta.extensions
                }
            ))
            raise e








import threading
from typing import Any, Coroutine
from arkitekt.actors.exceptions import ThreadedActorCancelled
from arkitekt.messages.postman.assign.assign_log import AssignLogMessage
import contextvars
from arkitekt.messages.postman.assign.assign_critical import AssignCriticalMessage
from arkitekt.messages.postman.assign.assign_cancelled import AssignCancelledMessage
from arkitekt.messages.postman.assign.assign_return import AssignReturnMessage
from arkitekt.messages.postman.assign.assign_yield import AssignYieldsMessage
from arkitekt.messages.postman.assign.assign_done import AssignDoneMessage
from arkitekt.messages.postman.provide.bounced_provide import BouncedProvideMessage
from arkitekt.threadvars import assign_message, transport, janus_queue, cancel_event
from arkitekt.messages.postman.assign.bounced_forwarded_assign import BouncedForwardedAssignMessage
from arkitekt.actors.base import Actor
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor
import asyncio
from arkitekt.packers.utils import shrink_inputs, shrink_outputs, expand_inputs, expand_outputs
import logging
import janus


logger = logging.getLogger(__name__)


class FunctionalActor(Actor):
    pass


    def __init__(self, *args, assign=None, on_provide=None, on_unprovide=None, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.assign = assign or self.assign
        self.on_provide = on_provide or self.on_provide
        self.on_unprovide = on_unprovide or self.on_unprovide


    def assign(self, *args, **kwargs):
        raise NotImplementedError("Please overwrite this")



class FunctionalFuncActor(FunctionalActor):
    assign: Coroutine[Any, Any, Any]

    
    async def progress(self, value, percentage):
        await self._progress(value, percentage)


    async def on_assign(self, message: BouncedForwardedAssignMessage):
        try:
            logger.info("Assigning Number two")
            args, kwargs = await expand_inputs(self.template.node, message.data.args, message.data.kwargs) if self.expand_inputs else (message.data.args, message.data.kwargs)
            
            transport.set(self.transport)
            assign_message.set(message)


            returns = await self.assign(*args, **kwargs)


            assign_message.set(None)
            transport.set(None)
            await self.transport.forward(AssignReturnMessage(data={
                "returns": await shrink_outputs(self.template.node, returns) if self.shrink_outputs else returns
            }, meta = {
                "reference": message.meta.reference,
                "extensions": message.meta.extensions
            }))


        except asyncio.CancelledError as e:

            await self.transport.forward(AssignCancelledMessage(data={
                "canceller": str(e)
            }, meta = {
                "reference": message.meta.reference,
                "extensions": message.meta.extensions
            }))



        except Exception as e:
            logger.exception(e)
            await self.transport.forward(AssignCriticalMessage(data={
                "type": e.__class__.__name__,
                "message": str(e)
            }, meta = {
                "reference": message.meta.reference,
                "extensions": message.meta.extensions
            }))




class FunctionalGenActor(FunctionalActor):


    async def progress(self, value, percentage):
        await self._progress(value, percentage)

    async def on_assign(self, message: BouncedForwardedAssignMessage):
        try:
            args, kwargs = await expand_inputs(self.template.node, message.data.args, message.data.kwargs) if self.expand_inputs else (message.data.args, message.data.kwargs)
            
            transport.set(self.transport)
            assign_message.set(message)

            async for returns in self.assign(*args, **kwargs):


                await self.transport.forward(AssignYieldsMessage(data={
                    "returns": await shrink_outputs(self.template.node, returns) if self.shrink_outputs else returns
                }, meta = {
                    "reference": message.meta.reference,
                    "extensions": message.meta.extensions
                }))


            assign_message.set(None)
            transport.set(None)


            await self.transport.forward(AssignDoneMessage(data={
                "returns": await shrink_outputs(self.template.node, returns) if self.shrink_outputs else returns
            }, meta = {
                "reference": message.meta.reference,
                "extensions": message.meta.extensions
            }))


        except asyncio.CancelledError as e:

            await self.transport.forward(AssignCancelledMessage(data={
                "canceller": str(e)
            }, meta = {
                "reference": message.meta.reference,
                "extensions": message.meta.extensions
            }))



        except Exception as e:
            logger.exception(e)
            await self.transport.forward(AssignCriticalMessage(data={
                "type": e.__class__.__name__,
                "message": str(e)
            }, meta = {
                "reference": message.meta.reference,
                "extensions": message.meta.extensions
            }))



class FunctionalThreadedFuncActor(FunctionalActor):

    def __init__(self, *args, nworkers=4, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.threadpool = ThreadPoolExecutor(nworkers)

    async def iterate_queue(self, async_q, message: BouncedForwardedAssignMessage):
        try:
            while True:
                val = await async_q.get()
                action = val[0]
                value = val[1]

                if action == "log":
                    message = AssignLogMessage(
                        data = {
                            "message": value[0],
                            "level": value[1]
                        },
                        meta = {**message.meta.dict(exclude={"type"})}
                    )

                    await self.transport.forward(message)
                    async_q.task_done()
                if action == "return":
                    await self.transport.forward(AssignReturnMessage(data={
                        "returns": await shrink_outputs(self.template.node, value) if self.shrink_outputs else value
                    }, meta = {**message.meta.dict(exclude={"type"})}
                    ))
                    async_q.task_done()
                    break
                if action == "exception":
                    async_q.task_done()
                    raise value 

            
        except asyncio.CancelledError as cancelled_error:
            logger.info(f"Received Cancellation to iterate over async queue")
            while True:
                val = await async_q.get()
                action = val[0]
                value = val[1]
                if action == "exception":
                    async_q.task_done()
                    try:
                        raise value
                    except ThreadedActorCancelled:
                        raise cancelled_error



    def _assign_threaded(self, queue, cancel_event_instance, message, args, kwargs):
        janus_queue.set(queue)
        assign_message.set(message)
        cancel_event.set(cancel_event_instance)
        try:
            result = self.assign(*args, **kwargs)
            queue.put(("return", result))
            queue.join()

        except Exception as e:
            logger.exception(e)
            queue.put(("exception", e))
            queue.join()

        janus_queue.set(None)
        assign_message.set(None)
        cancel_event.set(None)


    async def on_assign(self, message: BouncedForwardedAssignMessage):
        queue = janus.Queue()
        event = threading.Event()

        try:
            logger.info("Assigning Number two")
            args, kwargs = await expand_inputs(self.template.node, message.data.args, message.data.kwargs) if self.expand_inputs else (message.data.args, message.data.kwargs)
            

            threadedfut = self.loop.run_in_executor(self.threadpool, self._assign_threaded,  queue.sync_q,event,  message, args, kwargs)
            queuefut =  self.iterate_queue(queue.async_q, message)

            try:
                await asyncio.gather(asyncio.shield(threadedfut), asyncio.shield(queuefut))
                queue.close()
                await queue.wait_closed()

            except asyncio.CancelledError as e:
                await self.log("Received Cancellation for task")

                queuefut.cancel() # We cancel the quefuture and are now only waiting for cancellation requests
                event.set() # We are sending the request to the queue

                try:
                    await self.log("asdasdasd")
                    await queuefut
                except asyncio.CancelledError as e:
                    await self.log("Sucessfully Cancelled Thread")
                    raise e

        except asyncio.CancelledError as e:

            await self.transport.forward(AssignCancelledMessage(data={
                "canceller": str(e)
            }, meta = {
                "reference": message.meta.reference,
                "extensions": message.meta.extensions
            }))



        except Exception as e:
            logger.exception(e)
            await self.transport.forward(AssignCriticalMessage(data={
                "type": e.__class__.__name__,
                "message": str(e)
            }, meta = {
                "reference": message.meta.reference,
                "extensions": message.meta.extensions
            }))




class FunctionalThreadedGenActor(FunctionalActor):

    def __init__(self, *args, nworkers=4, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.threadpool = ThreadPoolExecutor(nworkers)


    
    async def iterate_queue(self, async_q, message: BouncedForwardedAssignMessage):
        try:
            while True:
                val = await async_q.get()
                action = val[0]
                value = val[1]

                if action == "log":
                    message = AssignLogMessage(
                        data = {
                            "message": value[0],
                            "level": value[1]
                        },
                        meta = {**message.meta.dict(exclude={"type"})}
                    )

                    await self.transport.forward(message)
                    async_q.task_done()
                if action == "yield":
                    await self.transport.forward(AssignYieldsMessage(data={
                        "returns": await shrink_outputs(self.template.node, value) if self.shrink_outputs else value
                    }, meta = {**message.meta.dict(exclude={"type"})}
                    ))
                    async_q.task_done()
                if action == "done":
                    await self.transport.forward(AssignDoneMessage(data={
                    }, meta = {**message.meta.dict(exclude={"type"})}
                    ))
                    async_q.task_done()
                    break
                if action == "exception":
                    async_q.task_done()
                    raise value

        except asyncio.CancelledError as cancelled_error:
            while True:
                val = await async_q.get()
                action = val[0]
                value = val[1]
                if action == "exception":
                    async_q.task_done()
                    try:
                        raise value
                    except ThreadedActorCancelled:
                        raise cancelled_error
            
                   



    def _assign_threaded(self, sync_queue, cancel_event_instance, message, args, kwargs):
        janus_queue.set(sync_queue)
        assign_message.set(message)
        cancel_event.set(cancel_event_instance)
        try:
            for result in self.assign(*args, **kwargs):
                sync_queue.put(("yield", result))
                sync_queue.join()

            sync_queue.put(("done","Happy doneness"))
            sync_queue.join()

        except Exception as e:
            logger.exception(e)
            sync_queue.put(("exception", e))
            sync_queue.join()

        janus_queue.set(None)
        assign_message.set(None) 
        cancel_event.set(None)


    async def on_assign(self, message: BouncedForwardedAssignMessage):
        queue = janus.Queue()
        event = threading.Event()
        
        try:
            logger.info("Assigning Number two")
            args, kwargs = await expand_inputs(self.template.node, message.data.args, message.data.kwargs) if self.expand_inputs else (message.data.args, message.data.kwargs)
            

            threadedfut = self.loop.run_in_executor(self.threadpool, self._assign_threaded,  queue.sync_q, event, message, args, kwargs)
            queuefut =  self.loop.create_task(self.iterate_queue(queue.async_q, message))

            try:
                await asyncio.gather(asyncio.shield(threadedfut), asyncio.shield(queuefut))
                queue.close()
                await queue.wait_closed()

            except asyncio.CancelledError as e:
                await self.log("Received Cancellation for task")

                queuefut.cancel() # We cancel the quefuture and are now only waiting for cancellation requests
                event.set() # We are sending the request to the queue

                try:
                    await self.log("asdasdasd")
                    await queuefut
                except asyncio.CancelledError as e:
                    await self.log("Sucessfully Cancelled Thread")

                    queue.close()
                    await queue.wait_closed()

                    raise e

        except asyncio.CancelledError as e:

            await self.transport.forward(AssignCancelledMessage(data={
                "canceller": str(e)
            }, meta = {
                "reference": message.meta.reference,
                "extensions": message.meta.extensions
            }))



        except Exception as e:
            logger.exception(e)
            await self.transport.forward(AssignCriticalMessage(data={
                "type": e.__class__.__name__,
                "message": str(e)
            }, meta = {
                "reference": message.meta.reference,
                "extensions": message.meta.extensions
            }))



from arkitekt.packers.structure import Structure
from typing import Callable
import inspect
import logging
import asyncio

logger = logging.getLogger(__name__)



class Transpiler():
    """A Transpiler registers for
    certain non native models and established
    a conversion algrohythm for an aritekt model
    """

    def __init__(self, type, structure, transpile: Callable, untranspile: Callable, loop=None, register=True) -> None:
        try:
            assert not issubclass(type, Structure), "Type cannot be a subclass of structure because structure are transpiled anyways"
            assert issubclass(structure, Structure), "structure must be a subclass of Structure in order to transpile to them"
        except Exception as e:
            raise Exception(f"Configuration Error: {type} , {structure}") from e
        self.type_name = type.__name__
        self.structure_name = structure.get_identifier()
        self.type = type
        self.structure = structure
        self.transpile_is_async = inspect.iscoroutinefunction(transpile)
        self._transpile = transpile

        self.untranspile_is_async = inspect.iscoroutinefunction(untranspile)
        self._untranspile = untranspile

        if register:

            from arkitekt.packers.transpilers.registry import get_transpiler_registry
            get_transpiler_registry().register_transpiler(self)
        

    async def transpile(self, object):
        logger.info(f"Transpiling from {self.type} to {self.structure}")
        if self.transpile_is_async: return await self._transpile(object)
        return await asyncio.get_event_loop().run_in_executor(None, self._transpile, object)

    async def untranspile(self, object):
        logger.info(f"Transpiling from {self.type} to {self.structure}")
        if self.untranspile_is_async: return await self._untranspile(object)
        return await asyncio.get_event_loop().run_in_executor(None, self._untranspile, object)






from abc import abstractmethod
from arkitekt.config import TransportProtocol
from arkitekt.messages.postman.assign.bounced_forwarded_assign import BouncedForwardedAssignMessage
from arkitekt.messages.postman.unassign.bounced_forwarded_unassign import BouncedForwardedUnassignMessage
from arkitekt.monitor.monitor import Monitor
from arkitekt.messages.postman.unprovide.bounced_unprovide import BouncedUnprovideMessage
from arkitekt.messages.postman.provide.bounced_provide import BouncedProvideMessage
from arkitekt.messages.postman.log import LogLevel
from arkitekt.messages.postman.reserve.bounced_reserve import BouncedReserveMessage
from arkitekt.transport.base import Transport
from arkitekt.messages.base import MessageDataModel, MessageModel
import asyncio
from typing import Callable, Dict, List, Tuple, Type
from arkitekt.transport.registry import TransportRegistry, get_current_transport_registry
from arkitekt.transport.websocket import WebsocketTransport
from arkitekt.schema.node import Node
from arkitekt.schema.template import Template
from arkitekt.ward import ArkitektConfig, ArkitektWard
from arkitekt.messages import ProvideLogMessage, ProvideCriticalMessage
from herre.console.context import get_current_console
from herre.herre import Herre, get_current_herre
from herre.wards.registry import get_ward_registry
from herre.wards.base import WardException
from arkitekt.legacy.utils import *
from arkitekt.registry import set_current_agent
from koil.koil import Koil, get_current_koil
from koil.loop import koil
from fakts import Fakts, get_current_fakts, Config

try:
    from rich.traceback import install
    install()
except:
    pass
import logging


logger = logging.getLogger(__name__)



class TemplateParams(dict):
    pass


async def parse_params(params: dict) -> TemplateParams:
    return TemplateParams(**params)


class AgentException(Exception):
    pass

class AgentConfig(Config):
    type: TransportProtocol = TransportProtocol.WEBSOCKET
    kwargs: dict = {}

    class Config:
        group = "arkitekt.agent"

    
class Agent:

    def __init__(self, *args, auto_login=True, auto_connect=True, loop=None, register=True,  with_monitor=False, herre: Herre = None, koil: Koil = None, fakts: Fakts = None, transport_registry: TransportRegistry= None, strict=False, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.auto_login = auto_login
        self.auto_connect = auto_connect
        self.strict = strict
        self.herre = herre or get_current_herre() 
        self.koil = koil or get_current_koil()
        self.fakts = fakts or get_current_fakts()
        self.loop = self.koil.loop
        
        self.ward: ArkitektWard = get_ward_registry().get_ward_instance("arkitekt")
        self.monitor = Monitor("Agent") if with_monitor else None
        self.panel = self.monitor.create_agent_panel(self) if self.monitor else None
        self.message_queue = asyncio.Queue()
        self.transport: Transport = None
        self.transport_registry = transport_registry or get_current_transport_registry()


        if register:
            set_current_agent(self)


    @abstractmethod
    async def handle_bounced_provide(self, bounced_provide: BouncedProvideMessage):
        raise NotImplementedError()

    @abstractmethod
    async def handle_bounced_unprovide(self, bounced_unprovide: BouncedUnprovideMessage):
        raise NotImplementedError()


    @abstractmethod
    async def handle_bounced_assign(self, bounced_assign: BouncedForwardedAssignMessage):
        raise NotImplementedError()

    @abstractmethod
    async def handle_bounced_unassign(self, bounced_unassign: BouncedForwardedUnassignMessage):
        raise NotImplementedError()



    async def broadcast(self, message: MessageModel):

        logger.info(f"Provider: [green] Received message {message}")
        try:
            if isinstance(message, BouncedProvideMessage):
                logger.info("Received Provide Request")
                assert message.data.template is not None, "Received Provision that had no Template???"
                await self.handle_bounced_provide(message)

            elif isinstance(message, BouncedUnprovideMessage):
                logger.info("Received Unprovide Request")
                assert message.data.provision is not None, "Received Unprovision that had no Provision???"
                await self.handle_bounced_unprovide(message)

            elif isinstance(message, BouncedForwardedAssignMessage):
                logger.info("Received Assing Request")
                assert message.data.provision is not None, "Received Assignment that had no Provision???"
                await self.handle_bounced_assign(message)

            elif isinstance(message, BouncedForwardedUnassignMessage):
                logger.info("Received Unassign Request")
                assert message.data.provision is not None, "Received Assignment that had no Provision???"
                await self.handle_bounced_unassign(message)

            else: 
                logger.info(f"Provider: [red] Received unknown message {message}")
        except Exception as e:
            logger.exception(e)


    

    async def on_transport_connected(self):
        return None

    async def on_transport_about_to_connect(self):
        return None
   

    async def aprovide(self):
        caused_ward_connect = False
        try:
            if not self.fakts.load_group:
                await self.fakts.aload()

            if not self.herre.logged_in:
                await self.herre.alogin()

            if not self.ward.connected:
                assert self.auto_connect, "We have not connected to Arkitekt before and autoconnect was set to false. Please connect a ward before or set auto_connect to True"
                await self.ward.aconnect()
                caused_ward_connect = True


            self.config = AgentConfig.from_fakts(fakts=self.fakts)
            self.transcript = self.ward.transcript
            self.transport = self.transport_registry.get_transport_for_protocol(self.config.type)(self.config.kwargs, broadcast=self.broadcast)


            await self.on_transport_about_to_connect()
            await self.transport.aconnect()
            await self.on_transport_connected()



            while True:
                await asyncio.sleep(1)

        except asyncio.CancelledError as e:

            if self.transport: await self.transport.adisconnect()
            if caused_ward_connect: await self.ward.adisconnect()
            raise e

        except Exception as e:
            logger.exception(e)
            raise e


    def provide(self, **kwargs):


        with get_current_console().status("Providing"):
            return koil(self.aprovide(), **kwargs)



